#
# COPYRIGHT (C) 2002-2011 Rajgopal Srinivasan (rajgopal.srinivasan@tcs.com)
#
"""
.. module:: annotateRegion
    :platform: Unix, Windows, MacOSX
    :synopsis: Region Annotation

.. moduleauthor:: Kunal Kundu (kunal.kundu@tcs.com)

For a given genome coordinate this module identifies its region i.e.
Intergenic, Intronic, Exonic, 5'UTR or 3'UTR.
"""


class RegionAnnotation():

    def __init__(self, spos, epos, ref, alt, refgene_entry):
        self.ref = ref
        self.alt = alt
        self.frames = refgene_entry.frames
        self.strand = refgene_entry.strand
        self.refseq_acc = refgene_entry.refseq_acc
        self.start_trns_pos = int(refgene_entry.str_transcript)
        self.end_trns_pos = int(refgene_entry.end_transcript)
        self.cds_start = int(refgene_entry.str_cds)
        self.cds_end = int(refgene_entry.end_cds)
        self.exon_str = [int(e) for e in refgene_entry.str_exon_pos.split(',')]
        self.exon_end = [int(e) for e in refgene_entry.end_exon_pos.split(',')]

        if self.strand == '+':
            self.spos = spos - 1
            self.epos = epos - 1
        else:
            self.spos = spos
            self.epos = epos

        # For SNP, INDELs
        self.poslist = range(self.spos, self.epos + 1)

        # For MNPs
        if len(self.ref) > 1 and len(self.alt) > 1:
            self.poslist = [ele[0] for ele in zip(self.poslist,
                                                  self.ref, self.alt)
                            if ele[1] != ele[2]]

        self.splice_site_range = 2
        self.region = None
        self.splice_site = None
        self.exons = None
        self.cdna_poslist = []

    def annotate_positive_strand_cds(self, position):
        region = None
        exon_num = 0
        cdna_pos = 0
        cdna_flag = False
        for x, y in zip(self.exon_str, self.exon_end):
            exon_num += 1
            if y < self.cds_start:
                exon_flag = False
                if x <= position < y:
                    region = 'UTR5'
                    cdna_pos += position - x + 1
                    cdna_flag = True
                    break
            elif x <= self.cds_start < y:
                exon_flag = True
                if x <= position < self.cds_start:
                    region = 'UTR5'
                    cdna_pos += position - x + 1
                    cdna_flag = True
                    break
                elif self.cds_start <= position < self.cds_end \
                                                    and position < y:
                    region = 'CodingExonic'
                    cdna_pos += position - x + 1
                    cdna_flag = True
                    break
                elif self.cds_end <= position < y:
                    region = 'UTR3'
                    cdna_pos += position - x + 1
                    cdna_flag = True
                    break
            elif x <= self.cds_end < y:
                exon_flag = False
                if self.cds_end <= position < y:
                    region = 'UTR3'
                    cdna_pos += position - x + 1
                    cdna_flag = True
                    break
                elif x <= position < self.cds_end:
                    region = 'CodingExonic'
                    cdna_pos += position - x + 1
                    cdna_flag = True
                    break
            elif x > self.cds_end:
                if x <= position < y:
                    region = 'UTR3'
                    cdna_pos += position - x + 1
                    cdna_flag = True
                    break
            elif exon_flag is True:
                if x <= position < y:
                    region = 'CodingExonic'
                    cdna_pos += position - x + 1
                    cdna_flag = True
                    break
            cdna_pos += y - x
        if cdna_flag is False:
            cdna_pos = None
        return region, exon_num, cdna_pos

    def annotate_negative_strand_cds(self, position):
        region = None
        self.exon_str.reverse()
        self.exon_end.reverse()
        exon_num = 0
        cdna_pos = 0
        cdna_flag = False
        for x, y in zip(self.exon_end, self.exon_str):
            exon_num += 1
            if y > self.cds_end:
                exon_flag = False
                if x >= position > y:
                    region = 'UTR5'
                    cdna_pos += x - position + 1
                    cdna_flag = True
                    break
            elif x >= self.cds_end > y:
                exon_flag = True
                if x >= position > self.cds_end:
                    region = 'UTR5'
                    cdna_pos += x - position + 1
                    cdna_flag = True
                    break
                elif self.cds_end >= position > self.cds_start \
                                                    and position > y:
                    region = 'CodingExonic'
                    cdna_pos += x - position + 1
                    cdna_flag = True
                    break
                elif self.cds_start >= position > y:
                    region = 'UTR3'
                    cdna_pos += x - position + 1
                    cdna_flag = True
                    break
            elif x >= self.cds_start > y:
                exon_flag = False
                if self.cds_start >= position > y:
                    region = 'UTR3'
                    cdna_pos += x - position + 1
                    cdna_flag = True
                    break
                elif x >= position > self.cds_start:
                    region = 'CodingExonic'
                    cdna_pos += x - position + 1
                    cdna_flag = True
                    break
            elif x < self.cds_start:
                if x >= position > y:
                    region = 'UTR3'
                    cdna_pos += x - position + 1
                    cdna_flag = True
                    break
            elif exon_flag is True:
                if x >= position > y:
                    region = 'CodingExonic'
                    cdna_pos += x - position + 1
                    cdna_flag = True
                    break
            cdna_pos += x - y
        if cdna_flag is False:
            cdna_pos = None
        self.exon_str.reverse()
        self.exon_end.reverse()
        exon_num = len(self.exon_str) - (exon_num - 1)
        return region, exon_num, cdna_pos

    def annotate_splice_site(self):
        if self.region[-8:] == 'boundary':
            reg1, reg2, _ = self.region.split('_', 2)
            if reg1 in ['UTR5', 'CodingExonic', 'UTR3', 'NonCodingExonic'] \
                     and reg2 in ['CodingIntronic', 'NonCodingIntronic']:
                if self.strand == '+':
                    self.splice_site = 'SpliceDonor'
                elif self.strand == '-':
                    self.splice_site = 'SpliceAcceptor'
            elif reg1 in ['CodingIntronic', 'NonCodingIntronic'] and reg2 in \
                        ['UTR5', 'CodingExonic', 'UTR3', 'NonCodingExonic']:
                if self.strand == '+':
                    self.splice_site = 'SpliceAcceptor'
                elif self.strand == '-':
                    self.splice_site = 'SpliceDonor'

        elif self.region[-8:] == 'Intronic':
            exon_end = self.exon_end
            exon_str = self.exon_str
            intron_cnt = 0
            del exon_str[0]
            del exon_end[-1]
            for intr_start, intr_end in zip(exon_end, exon_str):
                intron_cnt += 1
                if self.strand == '+':
                    if intr_start <= self.poslist[0] < intr_end and \
                                    intr_start <= self.poslist[-1] < intr_end:
                        dnr_dist = self.poslist[0] - intr_start + 1
                        acpt_dist = intr_end - self.poslist[-1]
                        if dnr_dist <= self.splice_site_range:
                            dnr_dist = 0
                        else:
                            dnr_dist -= 2
                        if acpt_dist <= self.splice_site_range:
                            acpt_dist = 0
                        else:
                            acpt_dist -= 2
                        break
                elif self.strand == '-':
                    if intr_start < self.poslist[0] <= intr_end and \
                                    intr_start < self.poslist[-1] <= intr_end:
                        acpt_dist = self.poslist[0] - intr_start
                        dnr_dist = intr_end - self.poslist[-1] + 1
                        if dnr_dist <= self.splice_site_range:
                            dnr_dist = 0
                        else:
                            dnr_dist -= 2
                        if acpt_dist <= self.splice_site_range:
                            acpt_dist = 0
                        else:
                            acpt_dist -= 2
                            break
            if dnr_dist == 0:
                self.splice_site = 'SpliceDonor'
            elif acpt_dist == 0:
                self.splice_site = 'SpliceAcceptor'
            else:
                self.splice_site = \
                            'Exon%s_SpliceDonor_%s__Exon%s_SpliceAcceptor_%s' \
                            % (intron_cnt, dnr_dist, intron_cnt + 1, acpt_dist)

    def annotate_nc(self, position, strand):
        region = None
        exon_num = 0
        cdna_pos = 0
        if strand == '+':
            for x, y in zip(self.exon_end, self.exon_str):
                exon_num += 1
                if y <= position < x:
                    region = 'Exonic'
                    cdna_pos += position - y + 1
                    break
                cdna_pos += x - y
        elif strand == '-':
            self.exon_str.reverse()
            self.exon_end.reverse()
            for x, y in zip(self.exon_end, self.exon_str):
                exon_num += 1
                if y < position <= x:
                    region = 'Exonic'
                    cdna_pos += x - position + 1
                    break
                cdna_pos += x - y
            self.exon_str.reverse()
            self.exon_end.reverse()
            exon_num = len(self.exon_end) - (exon_num - 1)
        if region is None:
            region = 'Intronic'
            exon_num = None
            cdna_pos = 0
        return region, exon_num, cdna_pos

    def annotate(self):
        reglist = []
        exons = []
        for position in self.poslist:
            reg = None
            if self.strand == '+':
                if position < self.start_trns_pos or \
                                position >= self.end_trns_pos:
                    reg = 'Intergenic'
                    exon_num = None
                    cdna_pos = 0
                else:
                    if self.cds_start != self.cds_end:
                        reg, exon_num, cdna_pos = \
                                    self.annotate_positive_strand_cds(position)
            elif self.strand == '-':
                if position - 1 < self.start_trns_pos or \
                            position - 1 >= self.end_trns_pos:
                        reg = 'Intergenic'
                        exon_num = None
                        cdna_pos = 0
                else:
                    if self.cds_start != self.cds_end:
                        reg, exon_num, cdna_pos = \
                                    self.annotate_negative_strand_cds(position)
            if reg is None and (self.refseq_acc.startswith('NR_') or
                            (len(list(set(self.frames.split(',')))) == 1 and
                                list(set(self.frames.split(',')))[0] == '-1')):
                t, exon_num, cdna_pos = self.annotate_nc(position, self.strand)
                reg = 'NonCoding' + t
            if reg is None:
                reg = 'CodingIntronic'
                exon_num = None
                cdna_pos = 0

            if reg not in reglist:
                reglist.append(reg)
            if exon_num and exon_num not in exons:
                exons.append(exon_num)

            if cdna_pos:
                self.cdna_poslist.append(cdna_pos)

        self.cdna_poslist.sort()
        if len(reglist) > 1:
            self.region = "_".join(reglist) + '_boundary'
        else:
            self.region = reglist[0]
        if exons:
            exons = [str(e) for e in exons]
            self.exons = '__'.join(exons)
        self.annotate_splice_site()
